#include <stdexcept>
#include <vector>
#include <msp/fs/dir.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "builder.h"
#include "component.h"
#include "executable.h"
#include "exportdefinitions.h"
#include "externaltask.h"
#include "gnucompiler.h"
#include "gnulinker.h"
#include "importlibrary.h"
#include "installedfile.h"
#include "objectfile.h"
#include "sharedlibrary.h"
#include "sourcepackage.h"
#include "staticlibrary.h"

using namespace std;
using namespace Msp;

GnuLinker::GnuLinker(Builder &b, const Architecture &a):
	Tool(b, a, "LINK")
{
	input_suffixes.push_back(".o");
	input_suffixes.push_back(".a");

	processing_unit = COMPONENT;

	default_linker = new Linker(*this, "CC");
	cxx_linker = new Linker(*this, "CXX");
}

GnuLinker::~GnuLinker()
{
	delete default_linker;
	delete cxx_linker;
}

Target *GnuLinker::create_target(const list<Target *> &sources, const string &arg)
{
	if(sources.empty())
		throw invalid_argument("GnuLinker::create_target");
	list<ObjectFile *> objs;
	Linker *linker = default_linker;
	for(list<Target *>::const_iterator i=sources.begin(); i!=sources.end(); ++i)
	{
		if(ObjectFile *obj = dynamic_cast<ObjectFile *>(*i))
		{
			objs.push_back(obj);
			if(obj->get_tool()->get_tag()=="CXX")
				linker = cxx_linker;
		}
		else
			throw invalid_argument("GnuLinker::create_target");
	}

	const Component &comp = *objs.front()->get_component();
	Binary *bin = 0;
	if(arg=="shared")
	{
		SharedLibrary *shlib = new SharedLibrary(builder, comp, objs);
		if(architecture->get_system()=="windows")
		{
			Tool &dlltool = builder.get_toolchain().get_tool("DLL");
			dlltool.create_target(*shlib);
		}
		bin = shlib;
	}
	else
		bin = new Executable(builder, comp, objs);
	bin->set_tool(*linker);
	return bin;
}

Target *GnuLinker::create_install(Target &target) const
{
	if(SharedLibrary *shlib = dynamic_cast<SharedLibrary *>(&target))
	{
		Tool &copy = builder.get_toolchain().get_tool("CP");
		InstalledFile *inst_tgt = dynamic_cast<InstalledFile *>(copy.create_target(target));
		if(architecture->get_system()=="windows")
			builder.get_build_graph().add_installed_target(*shlib->get_import_library());
		else
		{
			const Pattern &pattern = architecture->get_shared_library_patterns().front();
			string link_name = pattern.apply(shlib->get_libname());
			if(link_name!=FS::basename(inst_tgt->get_path()))
				inst_tgt->set_symlink(link_name);
		}
		return inst_tgt;
	}
	else
		return 0;
}

void GnuLinker::do_prepare()
{
	bool path_found = false;
	const FS::Path &sysroot = build_info.sysroot;

	Tool &compiler = builder.get_toolchain().get_tool("CC");
	if(dynamic_cast<GnuCompiler *>(&compiler))
	{
		compiler.prepare();
		FileTarget *compiler_exe = compiler.get_executable();
		if(compiler_exe)
		{
			ExternalTask::Arguments argv;
			argv.push_back(compiler_exe->get_path().str());
			argv.push_back("-Wl,--verbose");
			argv.push_back("-nostdlib");
			if(!sysroot.empty())
				argv.push_back("--sysroot="+sysroot.str());

			builder.get_logger().log("auxcommands", format("Running %s", join(argv.begin(), argv.end())));
			try
			{
				string output = ExternalTask::run_and_capture_output(argv, FS::Path(), true);
				string::size_type start = 0;
				while(start<output.size())
				{
					string::size_type search_dir = output.find("SEARCH_DIR(\"", start);
					if(search_dir==string::npos)
						break;

					search_dir += 12;
					string::size_type end = output.find("\");", search_dir);
					if(end==string::npos)
						break;

					FS::Path path;
					if(!output.compare(search_dir, 2, "=/"))
					{
						search_dir += 2;
						if(sysroot.empty())
							path = "/";
						else
							path = sysroot;
					}

					path /= output.substr(search_dir, end-search_dir);
					builder.get_logger().log("tools", format("Got %s system path: %s", tag, path));
					system_path.push_back(path);
					path_found = true;

					start = end+3;
				}
			}
			catch(...)
			{ }
		}
	}

	if(!path_found)
	{
		builder.get_logger().log("tools", format("No %s system path found, using defaults", tag));
		if(!sysroot.empty())
			system_path.push_back(sysroot/"usr/lib");
		else if(architecture->is_native())
		{
			system_path.push_back("/lib");
			system_path.push_back("/usr/lib");
			if(architecture->match_name("pc-32-linux"))
			{
				system_path.push_back("/lib/i386-linux-gnu");
				system_path.push_back("/usr/lib/i386-linux-gnu");
			}
			else if(architecture->match_name("pc-64-linux"))
			{
				system_path.push_back("/lib/x86_64-linux-gnu");
				system_path.push_back("/usr/lib/x86_64-linux-gnu");
			}
		}
		else
			system_path.push_back("/usr/"+architecture->get_cross_prefix()+"/lib");
	}
}

Task *GnuLinker::run(const Target &) const
{
	throw logic_error("GnuLinker should not be run directly");
}


GnuLinker::Linker::Linker(GnuLinker &p, const string &ct):
	SubTool(p),
	compiler_tag(ct)
{
	if(compiler_tag=="CC")
		set_command("gcc", true);
	else if(compiler_tag=="CXX")
		set_command("g++", true);
	else
		throw invalid_argument("GnuLinker::Linker::Linker");
}

string GnuLinker::Linker::create_build_signature(const BuildInfo &binfo) const
{
	string result = FS::basename(executable->get_path());
	result += ',';
	if(binfo.libmode<=BuildInfo::STATIC)
		result += 't';
	else
		result += 'd';
	if(binfo.strip)
		result += 's';
	if(!binfo.libs.empty())
	{
		result += ",l";
		result += join(binfo.libs.begin(), binfo.libs.end(), ",l");
	}
	return result;
}

void GnuLinker::Linker::do_prepare()
{
	parent.prepare();
	build_info = parent.get_build_info();
	system_path = parent.get_system_path();

	Tool &compiler = builder.get_toolchain().get_tool(compiler_tag);
	if(dynamic_cast<GnuCompiler *>(&compiler))
	{
		compiler.prepare();
		executable = compiler.get_executable();
	}
}

Task *GnuLinker::Linker::run(const Target &target) const
{
	const Binary &bin = dynamic_cast<const Binary &>(target);

	vector<string> argv;
	argv.push_back(executable->get_path().str());

	FS::Path work_dir = bin.get_component()->get_package().get_source_directory();

	if(const SharedLibrary *shlib = dynamic_cast<const SharedLibrary *>(&bin))
	{
		argv.push_back("-shared");
		argv.push_back("-fPIC");
		if(architecture->get_system()!="windows" && !shlib->get_soname().empty())
		{
			if(architecture->get_system()=="darwin")
			{
				argv.push_back("-install_name");
				argv.push_back(shlib->get_soname());

				const string &ver = shlib->get_package()->get_version();
				const string &if_ver = shlib->get_package()->get_interface_version();
				if(!ver.empty() && !if_ver.empty())
				{
					argv.push_back("-current_version");
					argv.push_back(ver);
					argv.push_back("-compatibility_version");
					argv.push_back(if_ver);
				}
			}
			else
				argv.push_back("-Wl,-soname,"+shlib->get_soname());
		}
	}

	BuildInfo binfo;
	target.collect_build_info(binfo);

	const FS::Path &sysroot = binfo.sysroot;
	if(!sysroot.empty())
		argv.push_back("--sysroot="+sysroot.str());

	FS::Path lib_dir = builder.get_prefix()/"lib";
	if(binfo.rpath_mode==BuildInfo::ABSOLUTE)
		argv.push_back("-Wl,-rpath,"+lib_dir.str());
	else
	{
		if(binfo.rpath_mode==BuildInfo::RELATIVE)
			argv.push_back("-Wl,-rpath,$ORIGIN/../lib");
		argv.push_back("-Wl,-rpath-link,"+lib_dir.str());
	}

	for(BuildInfo::PathList::const_iterator i=binfo.libpath.begin(); i!=binfo.libpath.end(); ++i)
		argv.push_back("-L"+i->str());
	if(binfo.strip)
		argv.push_back("-s");
	if(binfo.threads && architecture->get_system()!="windows" && architecture->get_system()!="darwin")
		argv.push_back("-pthread");

	const Architecture &native_arch = builder.get_native_arch();
	if(architecture->is_native() && architecture->get_bits()!=native_arch.get_bits())
		argv.push_back(format("-m%d", architecture->get_bits()));

	argv.push_back("-o");
	argv.push_back(relative(bin.get_path(), work_dir).str());

	for(BuildInfo::WordList::const_iterator i=binfo.keep_symbols.begin(); i!=binfo.keep_symbols.end(); ++i)
		argv.push_back("-u"+*i);

	bool static_link_ok = (binfo.libmode<=BuildInfo::STATIC);

	const Target::Dependencies &depends = target.get_dependencies();
	for(Target::Dependencies::const_iterator i=depends.begin(); i!=depends.end(); ++i)
	{
		FileTarget *file = dynamic_cast<FileTarget *>(*i);
		Target *tgt = (*i)->get_real_target();

		if(ObjectFile *obj = dynamic_cast<ObjectFile *>(tgt))
			argv.push_back(relative(obj->get_path(), work_dir).str());
		else if(StaticLibrary *stlib = dynamic_cast<StaticLibrary *>(tgt))
			argv.push_back((file?file:stlib)->get_path().str());
		else if(SharedLibrary *shlib = dynamic_cast<SharedLibrary *>(tgt))
		{
			argv.push_back("-l"+shlib->get_libname());
			static_link_ok = false;
		}
		else if(ImportLibrary *imp = dynamic_cast<ImportLibrary *>(tgt))
		{
			shlib = imp->get_shared_library();
			if(shlib)
				argv.push_back("-l"+shlib->get_libname());
			else
				argv.push_back((file?file:imp)->get_path().str());
			static_link_ok = false;
		}
	}

	for(BuildInfo::WordList::const_iterator i=binfo.libs.begin(); i!=binfo.libs.end(); ++i)
		if(i->size()>10 && !i->compare(i->size()-10, 10, ".framework"))
		{
			argv.push_back("-framework");
			argv.push_back(i->substr(0, i->size()-10));
		}

	if(static_link_ok)
		argv.push_back("-static");
	else
	{
		if(compiler_tag=="CXX")
		{
			BuildInfo::LibModeMap::const_iterator i = binfo.libmodes.find("stdc++");
			if(i!=binfo.libmodes.end() && i->second<=BuildInfo::STATIC)
				argv.push_back("-static-libstdc++");
		}

		if(architecture->get_system()=="windows")
			argv.push_back("-Wl,--enable-auto-import");
	}

	return new ExternalTask(argv, work_dir);
}
